import boto3
import boto3_session_cache
import botocore.session

from . import config


def session():
    aws_config = config.get('aws') or {}
    profile = aws_config.pop('profile_name', None)
    botocore_session = botocore.session.Session(profile=profile)
    resolver = botocore_session.get_component('credential_provider')
    provider = resolver.get_provider('assume-role')
    provider.cache = boto3_session_cache.JSONFileCache()
    return boto3.Session(botocore_session=botocore_session, **aws_config)


def ec2():
    return session().client('ec2')


def ssm():
    return session().client('ssm')



import json
import sys


def json_exit(data):
    sys.stderr.write('ERROR:\n')
    if data:
        try:
            json.dump(data, sys.stderr, indent=2)
        except Exception:
            sys.stderr.write(data)
    sys.exit(1)


def string_exit(message):
    sys.stderr.write('ERROR:\n')
    sys.stderr.write(message)
    sys.stderr.write('\n')
    sys.exit(1)



import copy
import subprocess
import sys
import toml


_config = {}


def _exec(command):
    return subprocess.check_output(command, shell=True).strip()


def _exec_all(data):
    if isinstance(data, basestring):
        if data.startswith('$(') and data.endswith(')'):
            data = _exec(data[2:-1])
    elif isinstance(data, dict):
        for key, value in data.items():
            data[key] = _exec_all(value)
    elif isinstance(data, list):
        data = [_exec_all(item) for item in data]
    return data


def get(key):

    keys = key.split('.')
    value = _config
    for key in keys:
        value = value.get(key)
        if value is None:
            return None

    if isinstance(value, (dict, list)):
        value = copy.deepcopy(value)

    return _exec_all(value)


def load(path, **defaults):
    _config.update(defaults)
    try:
        with open(path) as config_file:
            data = toml.load(config_file)
    except IOError as error:
        sys.stderr.write('Error reading config: {}\n'.format(error))
        sys.exit(1)
    except Exception as error:
        sys.stderr.write('Error parsing config: {}\n'.format(error))
        sys.exit(1)
    else:
        _config.update(data)



from __future__ import print_function

import os

from . import config


def _get_hostname(instance):
    return instance.get('PublicIpAddress') or instance['PrivateIpAddress']


def connect(instance, bastion):

    command = ['ssh']

    if config.get('verbose'):
        command += ['-v']

    if bastion:
        command += ['-A', '-t', _get_hostname(bastion), 'ssh']

    command += [_get_hostname(instance)]

    if config.get('verbose'):
        print('Running {}'.format(' '.join(command)))

    os.execlp('ssh', *command)



import curses

from curses import panel

from . import config


class Menu(object):

    def __init__(self, title, instances, stdscreen):
        self.window = stdscreen.subwin(0, 0)
        self.window.timeout(1000)
        self.window.keypad(1)
        self.panel = panel.new_panel(self.window)
        self.panel.hide()
        panel.update_panels()

        self.title = title

        self.position = 0
        self.instances = instances

    def addstr(self, y, x, str, attr=None):
        if y + 1 < self.max_y and x + len(str) < self.max_x:
            self.window.addstr(y, x, str, attr)

    def navigate(self, n):
        self.position += n
        if self.position < 0:
            self.position = len(self.instances) - 1
        elif self.position >= len(self.instances):
            self.position = 0

    def display(self):

        self.panel.top()
        self.panel.show()
        self.window.clear()

        while True:

            self.window.refresh()
            curses.doupdate()

            self.max_y, self.max_x = self.window.getmaxyx()

            # Display the menu title.
            self.addstr(1, 2, self.title, curses.A_NORMAL)
            self.addstr(2, 2, '-' * len(self.title), curses.A_NORMAL)

            for index, instance in enumerate(self.instances):

                # Highlight the selected instance.
                if index == self.position:
                    mode = curses.A_REVERSE
                else:
                    mode = curses.A_NORMAL

                # Display the instance.
                msg = _get_label(instance)
                self.addstr(3 + index, 2, msg, mode)

                # Write out spaces to handle when a message becomes shorter.
                self.addstr(
                    3 + index,
                    2 + len(msg),
                    ' ' * 100,
                    curses.A_NORMAL,
                )

            # Because window.timeout was called,
            # this returns -1 if nothing was pressed.
            key = self.window.getch()

            if key in [curses.KEY_ENTER, ord('\n')]:
                return self.instances[self.position]
            elif key == curses.KEY_UP:
                self.navigate(-1)
            elif key == curses.KEY_DOWN:
                self.navigate(1)
            elif key in (81, 113):
                # Either q or Q was pressed.
                return None

        self.window.clear()
        self.panel.hide()
        panel.update_panels()
        curses.doupdate()


def _display(screen, instances):
    curses.curs_set(0)
    menu = Menu('EC2 Instances', instances, screen)
    return menu.display()


def _get_label(instance):

    result = []
    for field in config.get('display.fields') or []:
        value = instance
        for key in field.split('.'):
            value = value.get(key)
        if value:
            result.append(value)
    return ' '.join(result) or instance['InstanceId']


def choose_instance(instances, search):
    if search:
        instances = [inst for inst in instances if search.lower() in _get_label(inst).lower()]
    if len(instances) == 1 and search:
        return instances[0]
    return curses.wrapper(_display, instances)



from . import aws, config, errors, ssm


def _filter_instances(instances, rules):
    result = []
    for instance in instances:
        if _rules_pass(instance, rules):
            result.append(instance)
    return result


def _instance_sort_key(instance):
    result = []
    for field in config.get('display.sort') or ['InstanceId']:
        value = instance
        for key in field.split('.'):
            value = value.get(key)
            if not value:
                break
        result.append(value or '')
    return result


def _rules_pass(obj, rules):

    for key, expected_value in rules.items():

        if key not in obj:
            return False

        if isinstance(expected_value, dict):
            nested_rules = expected_value
            if not _rules_pass(obj[key], nested_rules):
                return False

        elif obj[key] != expected_value:
            return False

    return True


def find_bastion(instance, instances):

    bastions = instances

    bastion_filters = config.get('bastion.filter')
    if not bastion_filters:
        return None

    for bastion_filter in bastion_filters.values():
        bastions = _filter_instances(bastions, bastion_filter)

    for bastion in bastions:
        if bastion['InstanceId'] == instance['InstanceId']:
            return None

    if bastions:
        return bastions[0]
    else:
        errors.string_exit('Bastion not found')


def find_instances():

    ec2 = aws.ec2()

    response = ec2.describe_instances()
    if response['ResponseMetadata']['HTTPStatusCode'] != 200:
        errors.json_exit(response)

    instances = []
    for reservation in response['Reservations']:
        for instance in reservation['Instances']:

            tags = {}
            for tag in instance.get('Tags') or []:
                tags[tag['Key']] = tag['Value']
            instance['Tags'] = tags

            instances.append(instance)

    instances = _filter_instances(instances, config.get('filter.ec2'))

    if instances and config.get('filter.ssm'):

        for info in ssm.find_instances():
            for instance in instances:
                if instance['InstanceId'] == info['InstanceId']:
                    instance.update(info)
                    break

        instances = _filter_instances(instances, config.get('filter.ssm'))

    return sorted(instances, key=_instance_sort_key)



from __future__ import print_function

import argparse

from . import config, ec2, menu, ssm, ssh


parser = argparse.ArgumentParser(prog=__package__)
parser.add_argument('config', help='Configuration file')
parser.add_argument('search', nargs='?', help='Search string')
parser.add_argument('-v', '--verbose', action='store_true', help='Verbose mode')

args = parser.parse_args()

config.load(args.config, verbose=args.verbose)

try:
    instances = ec2.find_instances()
    if instances:
        instance = menu.choose_instance(instances, args.search)
        if instance:
            bastion = ec2.find_bastion(instance, instances)
            if config.get('ssm'):
                ssm.add_ssh_key(instance, bastion)
            ssh.connect(instance, bastion)
    else:
        print('No matching instances found')
except KeyboardInterrupt:
    pass



from __future__ import print_function

import time

from botocore.exceptions import ClientError

from . import aws, config, errors


def _send_run_command(instance_ids):

    if config.get('verbose'):
        print('Sending SSM command')

    ssm = aws.ssm()

    result = ssm.send_command(
        InstanceIds=instance_ids,
        DocumentName=config.get('ssm.document_name'),
        Parameters=config.get('ssm.parameters'),
    )

    try:
        command_id = result['Command']['CommandId']
    except Exception:
        errors.json_exit(result)

    return command_id


def _wait_for_command(instance_ids, command_id):

    ssm = aws.ssm()

    for instance_id in instance_ids:

        result = {'Status': 'Pending'}

        while result['Status'] in ('Pending', 'InProgress', 'Delayed'):
            time.sleep(0.25)
            try:
                result = ssm.get_command_invocation(
                    CommandId=command_id,
                    InstanceId=instance_id,
                )
            except ClientError:
                pass

        if result['Status'] != 'Success':
            errors.json_exit(result)


def add_ssh_key(instance, bastion):

    instance_ids = [instance['InstanceId']]
    if bastion:
        instance_ids.append(bastion['InstanceId'])

    command_id = _send_run_command(
        instance_ids=instance_ids,
    )

    _wait_for_command(
        instance_ids=instance_ids,
        command_id=command_id,
    )


def find_instances():

    ssm = aws.ssm()

    response = ssm.describe_instance_information()
    if response['ResponseMetadata']['HTTPStatusCode'] != 200:
        errors.json_exit(response)

    return response['InstanceInformationList']
